/*
 * lirccd v0.8 - Middle layer daemon between lircd and its clients
 * Copyright (C) 2003  Fredrik Tolf (fredrik@dolda2000.cjb.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <getopt.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <sys/uio.h>
#include <fcntl.h>

#include "lirccd.h"
#include "command.h"
#include "usercommands.h"

#define LIRC_DEVICE "/dev/lircd"
#define LIRC_BUFSIZE 256

struct clicmd
{
    struct clicmd *next;
    unsigned char *name;
    int (*handler)(unsigned char **words, struct client *client);
};

struct client *clients = NULL;
struct binding *bindings = NULL;
struct binding *bindstack[BIND_DEPTH];
int bindlevel;
struct child *children = NULL;
struct alarm *alarms = NULL;
struct clicmd *clicmds = NULL;
int serverfd, lircfd;
static unsigned char lircbuf[LIRC_BUFSIZE];
static int lircdlen = 0;
int running;
static struct client *capture = NULL;
static unsigned char buf[1024];

unsigned char *sprintf2(unsigned char *format, ...)
{
    va_list args;
    unsigned char *buf;
    int len;
    
    va_start(args, format);
    len = vsnprintf(NULL, 0, format, args);
    vsnprintf((buf = malloc(len + 1)), len + 1, format, args);
    va_end(args);
    return(buf);
}

static unsigned char *quoteword(unsigned char *word)
{
    int i;
    unsigned char *buf;
    int bufsize, len;
    
    if((buf = malloc(bufsize = ((len = (strlen(word) + 1)) + 16))) == NULL)
    {
	errno = ENOMEM;
	return(NULL);
    }
    strcpy(buf, word);
    for(i = 0; buf[i]; i++)
    {
	if((buf[i] == ' ') || (buf[i] == '\\'))
	{
	    if(++len >= bufsize)
	    {
		if((buf = realloc(buf, bufsize += 16)) == NULL)
		{
		    free(buf);
		    errno = ENOMEM;
		    return(NULL);
		}
	    }
	    memmove(buf + i + 1, buf + i, len - i);
	    buf[i++] = '\\';
	}
    }
    return(buf);
}

static unsigned char *wordcmp(unsigned char **words)
{
    unsigned char *buf, *buf2;
    int len;
    
    if((buf = malloc(1)) == NULL)
    {
	errno = ENOMEM;
	return(NULL);
    }
    *buf = 0;
    len = 1;
    for(; *words; words++)
    {
	buf2 = quoteword(*words);
	if((buf = realloc(buf, len += (strlen(buf2) + 1))) == NULL)
	{
	    free(buf);
	    free(buf2);
	    errno = ENOMEM;
	    return(NULL);
	}
	if(*buf)
	    strcat(buf, " ");
	strcat(buf, buf2);
	free(buf2);
    }
    return(buf);
}

static unsigned char **wordexp(unsigned char *line)
{
    unsigned char **words;
    unsigned char *buf, *p;
    int bufsize, len, numwords;
    
    if((words = malloc(sizeof(unsigned char *))) == NULL)
    {
	errno = ENOMEM;
	return(NULL);
    }
    *words = NULL;
    numwords = 0;
    if((buf = malloc(bufsize = 16)) == NULL)
    {
	free(words);
	errno = ENOMEM;
	return(NULL);
    }
    *buf = 0;
    len = 0;
    for(p = line; 1; p++)
    {
	if(!*p || (*p == ' '))
	{
	    buf[len++] = 0;
	    buf = realloc(buf, len);
	    words[numwords++] = buf;
	    if((words = realloc(words, sizeof(unsigned char *) * (numwords + 1))) == NULL)
	    {
		free(words);
		free(buf);
		errno = ENOMEM;
		return(NULL);
	    }
	    words[numwords] = NULL;
	    if((buf = malloc(bufsize = 16)) == NULL)
	    {
		free(words);
		errno = ENOMEM;
		return(NULL);
	    }
	    *buf = 0;
	    len = 0;
	    if(!*p)
		break;
	} else if(*p == '\\') {
	    if(*(p + 1))
	    {
		buf[len++] = *(++p);
		if(len == bufsize)
		{
		    if((buf = realloc(buf, bufsize += 16)) == NULL)
		    {
			free(words);
			free(buf);
			errno = ENOMEM;
			return(NULL);
		    }
		}
	    }
	} else {
	    buf[len++] = *p;
	    if(len == bufsize)
	    {
		if((buf = realloc(buf, bufsize += 16)) == NULL)
		{
		    free(words);
		    free(buf);
		    errno = ENOMEM;
		    return(NULL);
		}
	    }
	}
    }
    return(words);
}

static void freewords(unsigned char **words)
{
    unsigned char **p;
    
    for(p = words; *p; p++)
	free(*p);
    free(words);
}

/* No, this cannot send more than 16 words, but it's supposed to be
 * used in a controlled environment. Just don't do anything stupid
 * with it.
 */
void sendline(int fd, ...)
{
    int i;
    va_list args;
    unsigned char *words[16], *cw, *buf;
    struct iovec v[2];
    
    va_start(args, fd);
    i = 0;
    do
    {
	cw = va_arg(args, unsigned char *);
	words[i++] = cw;
    } while(cw != NULL);
    va_end(args);
    buf = wordcmp(words);
    v[0].iov_base = buf;
    v[0].iov_len = strlen(buf);
    v[1].iov_base = "\n";
    v[1].iov_len = 1;
    writev(fd, v, 2);
    free(buf);
}

void sighandler(int signum)
{
    pid_t pid;
    struct child *cur;
    
    switch(signum)
    {
    case SIGCHLD:
	pid = waitpid(-1, NULL, WNOHANG);
	for(cur = children; cur != NULL; cur = cur->next)
	{
	    if(cur->pid == pid)
	    {
		cur->stopped = 1;
		break;
	    }
	}
	break;
    case SIGHUP:
    case SIGINT:
    case SIGTERM:
	running = 0;
	break;
    }
}

void freechild(struct child *child)
{
    struct child *cur;
    
    if(child == children)
    {
	children = child->next;
    } else {
	for(cur = children; cur->next != child; cur = cur->next);
	cur->next = child->next;
    }
    if(child->notify != NULL)
	free(child->notify);
    free(child->desc);
    free(child);
}

struct child *getchild(int id)
{
    struct child *cur;
    
    for(cur = children; cur != NULL; cur = cur->next)
    {
	if(cur->id == id)
	    return(cur);
    }
    return(NULL);
}

struct child *regchild(unsigned char *desc, int pid)
{
    static int curid = 0;
    struct child *new;
    
    new = (struct child *)malloc(sizeof(struct child));
    new->id = curid++;
    new->desc = strcpy(malloc(strlen(desc) + 1), desc);
    new->pid = pid;
    new->running = 1;
    new->stopped = 0;
    new->next = children;
    new->notify = NULL;
    children = new;
    return(new);
}

void killchildren(void)
{
    struct child *cur;
    
    for(cur = children; cur != NULL; cur = cur->next)
    {
	if(cur->running)
	    kill(cur->pid, SIGHUP);
    }
}

void checkchildren(void)
{
    struct child *cur;
    
    do
    {
	for(cur = children; cur != NULL; cur = cur->next)
	{
	    if(cur->running && cur->stopped)
	    {
		cur->running = 0;
		cur->stopped = 0;
		if(cur->notify != NULL)
		{
		    relval(runfunc(cur->notify, 1, valmakeint(newval(), cur->id), NULL));
		    break;
		}
	    }
	}
    } while(cur != NULL);
}

void freealarm(struct alarm *alarm)
{
    struct alarm *cur;
    
    if(alarm == alarms)
    {
	alarms = alarm->next;
    } else {
	for(cur = alarms; cur->next != alarm; cur = cur->next);
	cur->next = alarm->next;
    }
    free(alarm->func);
    free(alarm);
}

struct alarm *regalarm(time_t time, unsigned char *func)
{
    static int curid = 0;
    struct alarm *new;
    
    new = (struct alarm *)malloc(sizeof(struct alarm));
    new->id = curid++;
    new->time = time;
    new->func = strcpy(malloc(strlen(func) + 1), func);
    new->next = alarms;
    alarms = new;
    return(new);
}

void checkalarms(void)
{
    struct alarm *cur, *next;
    time_t curtime;
    
    curtime = time(NULL);
    for(cur = alarms; cur != NULL; cur = next)
    {
	next = cur->next;
	if(curtime >= cur->time)
	{
	    relval(runfunc(cur->func, 1, valmakeint(newval(), cur->id), NULL));
	    freealarm(cur);
	}
    }
}

struct client *newclient(int fd)
{
    struct client *new;
    
    new = (struct client *)malloc(sizeof(struct client));
    new->fd = fd;
    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
    if((new->buf = malloc(new->buflen = 64)) == NULL)
    {
	free(new);
	return(NULL);
    }
    new->datalen = 0;
    new->appname = NULL;
    new->next = clients;
    new->destroy = 0;
    new->sendraw = 0;
    clients = new;
    return(new);
}

void freeclient(struct client *client)
{
    struct client *cur;
    
    if(capture == client)
	capture = NULL;
    if(client == clients)
    {
	clients = client->next;
    } else {
	for(cur = clients; cur->next != client; cur = cur->next);
	cur->next = client->next;
    }
    close(client->fd);
    if(client->appname != NULL)
	free(client->appname);
    free(client->buf);
    free(client);
}


void addbinding(struct binding **dest, struct binding *binding)
{
    struct binding *cur;
    
    if(*dest == NULL)
    {
	*dest = binding;
	return;
    }
    for(cur = *dest; cur->next != NULL; cur = cur->next);
    cur->next = binding;
}

struct binding *newbinding(unsigned char *remote, unsigned char *button, struct instruction *code)
{
    struct binding *new;
    int ret;
    
    new = (struct binding *)malloc(sizeof(struct binding));
    new->next = NULL;
    new->child = NULL;
    new->code = code;
    new->flags = 0;
    if(ret = regcomp(&new->remote, remote, REG_EXTENDED | REG_NOSUB | REG_ICASE))
    {
	regerror(ret, &new->remote, buf, 1024);
	fprintf(stderr, "%s\n", buf);
	exit(1);
    }
    if(ret = regcomp(&new->button, button, REG_EXTENDED | REG_NOSUB | REG_ICASE))
    {
	regerror(ret, &new->remote, buf, 1024);
	fprintf(stderr, "%s\n", buf);
	exit(1);
    }
    return(new);
}

void regclicmd(unsigned char *name, int (*handler)(unsigned char **words, struct client *client))
{
    struct clicmd *cmd;
    
    cmd = malloc(sizeof(*cmd));
    cmd->name = strdup(name);
    cmd->handler = handler;
    cmd->next = clicmds;
    clicmds = cmd;
}

void handlebutton(unsigned char *code, int rep, unsigned char *button, unsigned char *remote)
{
    int i, r, f, done;
    struct binding **stackcopy, *cur;
    struct value *v;
    
    v = valmakestr(newval(), code);
    relval(storeval(getsym("code"), v));
    relval(v);
    v = valmakestr(newval(), button);
    relval(storeval(getsym("button"), v));
    relval(v);
    v = valmakestr(newval(), remote);
    relval(storeval(getsym("remote"), v));
    relval(v);
    v = valmakeint(newval(), rep);
    relval(storeval(getsym("repeat"), v));
    relval(v);
    stackcopy = malloc(sizeof(*stackcopy) * (bindlevel));
    memcpy(stackcopy, bindstack, sizeof(*stackcopy) * (bindlevel));
    r = 1;
    f = 1;
    done = 0;
    for(i = bindlevel - 1; i >= 0; i--)
    {
	for(cur = stackcopy[i]; cur != NULL; cur = cur->next)
	{
	    if(r || ((cur->flags & BIND_RECURSIVE) && (f || !(cur->flags & BIND_FALLBACK))))
	    {
		if(!regexec(&cur->remote, remote, 0, NULL, 0) && !regexec(&cur->button, button, 0, NULL, 0))
		{
		    if(cur->code != NULL)
		    {
			run(cur->code);
			f = 0;
			checkstack();
		    }
		    if(cur->child != NULL)
		    {
			if(bindlevel < BIND_DEPTH)
			{
			    bindstack[bindlevel++] = cur->child;
			    break;
			} else {
			    fprintf(stderr, "Attempting to exceed bind stack depth\n");
			}
		    }
		    if(cur->flags & BIND_STOP)
		    {
			done = 1;
			break;
		    }
		}
	    }
	}
	r = 0;
	if(done)
	    break;
    }
    free(stackcopy);
}

static int cc_ver(unsigned char **words, struct client *client)
{
    /* Note: This returns the protocol version, not the program version */
    sendline(client->fd, "repl", "101", "1", NULL);
}

static int cc_ident(unsigned char **words, struct client *client)
{
    if(words[1] == NULL)
    {
	sendline(client->fd, "repl", "301", "Wrong number of arguments", NULL);
	return(0);
    }
    if(client->appname != NULL)
	free(client->appname);
    if((client->appname = strdup(words[1])) == NULL)
    {
	client->destroy = 1;
	return(1);
    }
    sendline(client->fd, "repl", "100", "Success", NULL);
    return(0);
}

static int cc_raw(unsigned char **words, struct client *client)
{
    if(words[1] == NULL)
    {
	sendline(client->fd, "repl", "301", "Wrong number of arguments", NULL);
	return(0);
    }
    if(client->sendraw = atoi(words[1]))
    {
	sendline(client->fd, "repl", "100", "Sending raw keypresses", NULL);
    } else {
	sendline(client->fd, "repl", "100", "Not sending raw keypresses", NULL);
    }
    return(0);
}

static int cc_cap(unsigned char **words, struct client *client)
{
    if(words[1] == NULL)
    {
	sendline(client->fd, "repl", "301", "Wrong number of arguments", NULL);
	return(0);
    }
    if(atoi(words[1]))
    {
	if(capture == NULL)
	{
	    capture = client;
	    sendline(client->fd, "repl", "100", "Captured", NULL);
	} else {
	    if(capture->appname != NULL)
		sendline(client->fd, "repl", "400", capture->appname, "Capture in use", NULL);
	    else
		sendline(client->fd, "repl", "401", "Capture in use", NULL);
	}
    } else {
	if(capture == client)
	{
	    capture = NULL;
	    sendline(client->fd, "repl", "100", "Released", NULL);
	} else {
	    if(capture->appname != NULL)
		sendline(client->fd, "repl", "400", capture->appname, "Capture in use", NULL);
	    else
		sendline(client->fd, "repl", "401", "Capture in use", NULL);
	}
    }
}

static void regclicommands(void)
{
    regclicmd("ver", cc_ver);
    regclicmd("ident", cc_ident);
    regclicmd("raw", cc_raw);
    regclicmd("cap", cc_cap);
}

static void parsecodes(int clientmode)
{
    struct client *client, *nclient;
    int newfd;
    fd_set rfds;
    struct timeval tv;
    unsigned char *p;
    int ret;
    int repeat;
    unsigned char *remote, *button, *code;
    unsigned char **words;
    struct clicmd *clicmd;
    
    running = 1;
    remote = malloc(128);
    button = malloc(128);
    code = malloc(128);
    while(running)
    {
	FD_ZERO(&rfds);
	FD_SET(lircfd, &rfds);
	FD_SET(serverfd, &rfds);
	for(client = clients; client != NULL; client = client->next)
	    FD_SET(client->fd, &rfds);
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if(select(FD_SETSIZE, &rfds, NULL, NULL, &tv) < 0)
	{
	    if(errno == EINTR)
	    {
		FD_ZERO(&rfds);
	    } else {
		perror("select");
		break;
	    }
	}
	checkalarms();
	checkchildren();
	if(FD_ISSET(lircfd, &rfds))
	{
	    ret = read(lircfd, lircbuf + lircdlen, LIRC_BUFSIZE - lircdlen);
	    if((ret < 0) && (errno != EINTR))
	    {
		perror("read");
		break;
	    }
	    if(ret == 0)
	    {
		fprintf(stderr, "Connection closed by lircd\n");
		break;
	    }
	    if(ret > 0)
	    {
		lircdlen += ret;
		while((p = memchr(lircbuf, '\n', lircdlen)) != NULL)
		{
		    *(p++) = 0;
		    sscanf(lircbuf, "%s %x %s %s", code, &repeat, button, remote);
		    memmove(lircbuf, p, lircdlen -= (p - lircbuf));
		    sprintf(buf, "%i", repeat);
		    if(capture != NULL)
		    {
			sendline(capture->fd, "bttn", code, buf, button, remote, NULL);
		    } else {
			for(client = clients; client != NULL; client = client->next)
			{
			    if(client->sendraw)
				sendline(client->fd, "bttn", code, buf, button, remote, NULL);
			}
			handlebutton(code, repeat, button, remote);
		    }
		}
		if(lircdlen == LIRC_BUFSIZE)
		    lircdlen = 0;
	    }
	}
	if(FD_ISSET(serverfd, &rfds))
	{
	    if((newfd = accept(serverfd, NULL, 0)) >= 0)
		newclient(newfd);
	}
	for(client = clients; client != NULL; client = nclient)
	{
	    nclient = client->next;
	    if(FD_ISSET(client->fd, &rfds))
	    {
		ret = read(client->fd, client->buf + client->datalen, client->buflen - client->datalen);
		if(((ret < 0) && (errno != EINTR)) || (ret == 0))
		{
		    freeclient(client);
		    if((clients == NULL) && clientmode)
			running = 0;
		    continue;
		}
		client->datalen += ret;
		while((p = memchr(client->buf, '\n', client->datalen)) != NULL)
		{
		    *(p++) = 0;
		    words = wordexp(client->buf);
		    ret = 0;
		    for(clicmd = clicmds; clicmd != NULL; clicmd = clicmd->next)
		    {
			if(!strcmp(words[0], clicmd->name))
			{
			    ret = 1;
			    if(clicmd->handler(words, client))
				break;
			}
		    }
		    if(!ret)
			sendline(client->fd, "repl", "300", "Command not known", NULL);
		    freewords(words);
		    memmove(client->buf, p, client->datalen -= (p - client->buf));
		}
		if(client->datalen == client->buflen)
		{
		    if(client->buflen < 1024)
		    {
			client->buf = realloc(client->buf, client->buflen += 64);
		    } else {
			freeclient(client);
			continue;
		    }
		}
		if(client->destroy)
		    freeclient(client);
	    }
	}
    }
    free(button);
    free(remote);
    free(code);
}

int main(int argc, unsigned char **argv)
{
    int o;
    int status;
    unsigned char *homedir, *sockname, *configfile;
    int nofork, clientmode;
    FILE *config;
    pid_t cpp;
    int pipefd[2];
    struct stat statb;
    struct value *v;
    struct instruction *init;
    struct sockaddr_un name;
    
    homedir = getpwuid(getuid())->pw_dir;
    sockname = sprintf2("%s/.lircc", homedir);
    configfile = sprintf2("%s/.lircrc", homedir);
    nofork = 0;
    clientmode = 0;
    while((o = getopt(argc, (char **)argv, "hC:s:fc")) >= 0)
    {
	switch(o)
	{
	case 'C':
	    free(configfile);
	    configfile = strdup(optarg);
	    break;
	case 's':
	    free(sockname);
	    sockname = strdup(optarg);
	    break;
	case 'f':
	    nofork = 1;
	    break;
	case 'c':
	    clientmode = 1;
	    break;
	case 'h':
	case '?':
	case ':':
	default:
	    fprintf(stderr, "usage: lirccd [-hcf] [-C configfile] [-s sockfile]\n");
	    exit((o == 'h')?0:1);
	}
    }
    initcommand();
    regusercommands();
    regclicommands();
    if(pipe(pipefd))
    {
	perror("pipe");
	return(1);
    }
    if((cpp = fork()) < 0)
    {
	perror("fork");
	return(1);
    }
    if(!cpp)
    {
	close(pipefd[0]);
	dup2(pipefd[1], 1);
	if(pipefd[1] != 1)
	    close(pipefd[1]);
	execlp("cpp", "cpp", configfile, NULL);
	exit(127);
    }
    free(configfile);
    close(pipefd[1]);
    config = fdopen(pipefd[0], "r");
    yyrestart(config);
    yyparse();
    fclose(config);
    if(waitpid(cpp, &status, 0) < 0)
    {
	perror("waitpid");
	return(1);
    }
    if(status)
    {
	fprintf(stderr, "cpp terminated abnormally (%i)\n", status);
	return(1);
    }
    v = newval();
    newsym("code", v);
    relval(v);
    v = newval();
    newsym("button", v);
    relval(v);
    v = newval();
    newsym("remote", v);
    relval(v);
    v = newval();
    newsym("repeat", v);
    relval(v);
    signal(SIGCHLD, sighandler);
    signal(SIGHUP, sighandler);
    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);
    if((init = getfunc("init")) != NULL)
    {
	run(init);
	pop();
	checkstack();
    }
    bindstack[0] = bindings;
    bindlevel = 1;
    srand(time(NULL));
    if(lstat(sockname, &statb) == 0)
    {
	if(!S_ISSOCK(statb.st_mode))
	{
	    fprintf(stderr, "%s is not a socket\n", sockname);
	    return(1);
	} else {
	    unlink(sockname);
	}
    } else {
	if(errno != ENOENT)
	{
	    perror(sockname);
	    return(1);
	}
    }
    if((serverfd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    {
	perror("socket");
	return(1);
    }
    name.sun_family = AF_UNIX;
    strcpy(name.sun_path, sockname);
    free(sockname);
    if(bind(serverfd, (struct sockaddr *)&name, sizeof(name)) < 0)
    {
	perror("could not bind server socket");
	exit(1);
    }
    if(listen(serverfd, 16) < 0)
    {
	perror("could not listen on server socket");
	exit(1);
    }
    if((lircfd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
    {
	perror("socket");
	return(1);
    }
    name.sun_family = AF_UNIX;
    strcpy(name.sun_path, LIRC_DEVICE);
    if(connect(lircfd, (struct sockaddr *)&name, sizeof(name)))
    {
	perror("could not connect to " LIRC_DEVICE);
	return(1);
    }
    if(!nofork)
	daemon(0, 0);
    parsecodes(clientmode);
    close(lircfd);
    close(serverfd);
    killchildren();
    return(0);
}
